@@ -24,77 +24,133 @@ use warnings;
 # All drivers are listed, most parameters have default values and
 # legal values written in comments in a uniform way. Hence this file
 # (and comments) can be parsed to retrieve the information required to
-# create a consistent model for LcdProc configuration.
+# create a consistent model for LcdProc configuration. Some useful
+# parameters are commented out in LCD.conf. So some processing is
+# required to be able to create a model with these commented
+# parameters.  See below for this processing.
 
-# This script performs 3 tasks:
-# 1/ parse LCDd.conf template
-# 2/ mine the information there and translate them in a format suitable to create
+# This script performs the following tasks:
+# 1/ check whether generating the model is necessary (or possible)
+# 2/ pre-process LCDd.conf template
+# 3/ parse the new LCDd.conf template
+# 4/ mine the information there and translate them in a format suitable to create
 #    a model. Comments are used to provide default and legal values and also to provide
 #    user documentation
-# 3/ Write the resulting LCDd model
+# 5/ Write the resulting LCDd model
 
 use Config::Model 2.076;
 use Config::Model::Itself 2.005;    # to create the model
-use Config::Model::Backend::IniFile;
 
 use 5.010;
-use IO::File;
-use IO::String;
+use Path::Tiny;
 use Getopt::Long;
 
-say "Building lcdproc model from upstream LCDd.conf file..." ;
-
 my $verbose = 0;
 my $show_model = 0;
-my $result = GetOptions ("verbose"  => \$verbose,
-    "model" => \$show_model);
+my $force = 0;
+my $result = GetOptions (
+    "verbose"  => \$verbose,
+    "model" => \$show_model,
+    "force" => \$force,
+);
 
 die "Unknown option. Expected -verbose or -show_model" unless $result ;
 
+########################
+#
+# Step 1: Check whether generating lcdproc model is necessary.
+
+my $target = "lib/Config/Model/models/LCDd.pl";
+my $script = "script/lcdconf2model.pl";
+my $source = "lcdproc/LCDd.conf" ;
+
+if (-e $target and -M $target < -M $script and -M $target < -M $source) {
+    say "LcdProc model is up to date";
+    exit unless $force;
+}
+
+say "Building lcdproc model from upstream LCDd.conf file..." ;
+
+###########################
+#
+# Step 2: pre-process LCDd.conf (INI file)
+
+# Here's the LCDd.conf pre-processing mentioned above
+
+# read LCDd.conf
+my $path = path('.');
+my @lines    = $path->child('lcdproc/LCDd.conf')->lines;
+
+# un-comment commented parameters and put value as default value
+foreach my $line (@lines) {
+    $line =~ s/^#(\w+)=(.*)/# [default: $2]\n$1=$2/;
+}
+
+# write pre-processed files
+my $tmp = $path->child('tmp');
+$tmp->mkpath;
+$tmp->child('LCDd.conf')->spew(@lines);
+
+###########################
+#
+# Step 3: parse LCDd.conf (INI file)
+
+# Problem: comments must also be retrieved and associated with INI
+# class and parameters
+
+# Fortunately, Config::Model::Backend::IniFile can already perform this
+# task.
+
+# On the other hand, Config::Model::Backend::IniFile must store its
+# values in a configuration tree. A model suitable for LCDd.conf that
+# accepts any INI class and any INI parameter must be created
+
 # Dump stack trace in case of error
 Config::Model::Exception::Any->Trace(1) ;
 
 # one model to rule them all
 my $model = Config::Model->new();
 
-###########################
-#
-# Step 1: parse LCDd.conf (INI file)
+# The model for pre-precessed LCDd.conf must be made of 2 classes:
+# - the main config class that contains INI class names (named Dummy here)
+# - the child class that contains data from a elements of the INI
+#   classes (named Dummy::Class)
 
-# Problem: comments must also be retrieved and associated with INI
-# class and parameters
-
-# Fortunately, Config::Model::Backend::IniFile can already perform this
-# task. But Config::Model::Backend::IniFile must store its values in a
-# configuration tree.
-
-# So let's create a model suitable for LCDd.conf that accepts any
-# INI class and any INI parameter
+# For techinical reason, the lower class (Dummy::Class) must be
+# created first.
 
 # The class is used to store any parameter found in an INI class
 $model->create_config_class(
     name   => 'Dummy::Class',
     accept => [
-        'Hello|GoodBye' => { type => 'list',
+        'Hello|GoodBye' => {
+            type => 'list',
             cargo => { qw/type  leaf value_type uniline/}
         },
-        '.*' => {qw/type leaf value_type uniline/},
+        '.*' => {
+            type => 'leaf',
+            value_type => 'uniline'
+        }
     ],
 );
 
-# Store any INI class, and use Dummy::Class to hold parameters. 
-
-# Note that a INI backend could be created here. But, some useful
-# parameters are commented out in LCD.conf. So some processing is
-# required to be able to create a model with these commented parameters.
-# See below for this processing.
-
+# This class contains any INI class, and use Dummy::Class to hold parameters.
 $model->create_config_class(
     name   => 'Dummy',
-    accept => [ '.*' => {qw/type node config_class_name Dummy::Class/}, ],
+    accept => [
+        '.*' => {
+            type => 'node',
+            config_class_name => 'Dummy::Class'
+        }
+    ],
+    read_config => [{
+        backend => 'IniFile',
+        config_dir => 'tmp', # created above
+        file => 'LCDd.conf'
+    }]
 );
 
-# Now the dummy configuration class is created. Let's create a 
+# Now the dummy configuration class is created. Let's create a
 # configuration tree to store the data from LCDd.conf
 
 my $dummy = $model->instance(
@@ -102,29 +158,9 @@ my $dummy = $model->instance(
     root_class_name => 'Dummy',
 )-> config_root;
 
-# read LCDd.conf
-my $lcd_file = IO::File->new('lcdproc/LCDd.conf');
-my @lines    = $lcd_file->getlines;
-
-# Here's the LCDd.conf pre-processing mentioned above
-
-# un-comment commented parameters and put value as default value
-foreach (@lines) { s/^#(\w+)=(.*)/# [default: $2]\n$1=$2/ }
-
-# pre-processing is done.
-
-# store the munged LCDd.conf in a IO::Handle usable by INI backend
-my $ioh = IO::String->new( join( '', @lines ) );
-
-# Create the INI backend
-my $ini_backend = Config::Model::Backend::IniFile->new( node => $dummy );
-
-# feed the munged LCDd.conf content into INI backend
-$ini_backend->read( io_handle => $ioh );
-
 ##############################################
 #
-# Step 2: Mine the LCDd.conf information and create a model
+# Step 4: Mine the LCDd.conf information and create a model
 #
 
 # Create a meta tree that will contain LCDd model
@@ -138,17 +174,13 @@ my $meta_root = $model->instance(
 $meta_root->grab("class:LCDd class_description")->store( $dummy->annotation );
 
 # append my own text
-my $extra_description =
-    "Model information extracted from template /etc/LCDd.conf"
-  . "\n\n=head1 BUGS\n\nThis model does not support to load several drivers. Loading "
-  . "several drivers is probably a marginal case. Please complain to the author if this "
-  . "assumption is false";
+my $extra_description = "Model information was extracted from /etc/LCDd.conf";
 $meta_root->load(qq!class:LCDd class_description.="\n\n$extra_description"!);
 
 # add legal stuff
 $meta_root->load( qq!
-    class:LCDd 
-        copyright:0="2011-2013, Dominique Dumont"
+    class:LCDd
+        copyright:0="2011-2016, Dominique Dumont"
         copyright:1="1999-2013, William Ferrell and others"
         license="GPL-2"
 !
@@ -156,19 +188,20 @@ $meta_root->load( qq!
 
 # add INI backend (So LCDd model will be able to read INI files)
 $meta_root->load( qq!
-    class:LCDd 
-        read_config:0 
-            backend=ini_file 
-            config_dir="/etc" 
+    class:LCDd
+        read_config:0
+            backend=ini_file
+            config_dir="/etc"
             file="LCDd.conf"
 !
 );
 
-# Note: all the load calls above could be done as one call. But I choose
-#       to split them for better clarity
+# Note: all the load calls above could be done in one call. They are
+# split in several class to clarify what's going on.
 
-
-# This array contains all INI classes found in LCDd.conf,
+# Now, let's use the information retrieved by /etc/LCDd.conf
+# and stored in Dummy tree.
+# @ini_classes array contains all INI classes found in LCDd.conf,
 # make sure to put server in first, and sort the rest
 my @ini_classes = sort grep { $_ ne 'server'} $dummy->get_element_name;
 unshift @ini_classes, 'server' ;
@@ -180,7 +213,7 @@ my %dispatch;
 # first create the default case which will be used for most parameters
 # This subs is passed: the INI class name, the INI parameter name
 # the comment attached to the parameter, the INI value, and an optional
-# value type  
+# value type
 $dispatch{_default_} = sub {
     my ( $ini_class, $ini_param, $info_r, $ini_v, $value_type ) = @_;
 
@@ -190,7 +223,7 @@ $dispatch{_default_} = sub {
 
     # get semantic information from comment (written between square brackets)
     my $square_model = '';
-    
+
     my $square_rexp = '\[(\s*\w+\s*:[^\]]*)\]';
     if ($$info_r =~ /$square_rexp/s) {
         my $info = $1 ;
@@ -198,7 +231,7 @@ $dispatch{_default_} = sub {
         $$info_r =~ s/$square_rexp//gs; # remove all remaining square_rexp
         $square_model .= ' '. info_to_model($info,$value_type, $info_r) ;
     }
-    
+
     unless ($square_model) {
         # or use the value found in INI file as default
         $ini_v =~ s/^"//g;
@@ -216,7 +249,7 @@ $dispatch{_default_} = sub {
             if $verbose;
         $$info_r =~ s/$curly_rexp//s;
     }
-    
+
     # return a string containing model specifications
     # spec in curly model may override spec in square model
     return $load . $square_model . $curly_model ;
@@ -235,30 +268,35 @@ $dispatch{"LCDd::server"}{Driver} = sub {
     return $load;
 };
 
-# Ensure that DriverPath will end with a slash
+# Ensure that DriverPath ends with a slash by adding a match clause
 $dispatch{"LCDd::server"}{DriverPath} = sub {
     return $dispatch{_default_}->( @_ ) . q! match="/$"! ;
 };
 
-# like default but ensure that parameter is integer
-$dispatch{"LCDd::server"}{WaitTime} = $dispatch{"LCDd::server"}{ReportLevel} =
- $dispatch{"LCDd::picolcd"}{LircFlushThreshold} = $dispatch{"LCDd::server"}{Port}   = sub {
-    my ( $class, $elt, $info_r, $ini_v ) = @_;
-    return $dispatch{_default_}->( @_, 'integer' );
-  };
+# like default but ensure that the parameter is integer
+$dispatch{"LCDd::server"}{WaitTime}
+    = $dispatch{"LCDd::server"}{ReportLevel}
+    = $dispatch{"LCDd::picolcd"}{LircFlushThreshold}
+    = $dispatch{"LCDd::server"}{Port}
+    = sub {
+        my ( $class, $elt, $info_r, $ini_v ) = @_;
+        return $dispatch{_default_}->( @_, 'integer' );
+    };
 
 # special dispatch case
 my %override ;
 
 # Handle display content
-$override{"LCDd::server"}{GoodBye} = $override{"LCDd::server"}{Hello} = sub {
-    my ( $class, $elt ) = @_;
-    my $ret = qq( class:"$class" element:$elt type=list ) ;
-    $ret .= 'cargo type=leaf value_type=uniline';
-    return $ret ;
-};
+$override{"LCDd::server"}{GoodBye}
+    = $override{"LCDd::server"}{Hello}
+    = sub {
+        my ( $class, $elt ) = @_;
+        my $ret = qq( class:"$class" element:$elt type=list ) ;
+        $ret .= 'cargo type=leaf value_type=uniline';
+        return $ret ;
+    };
 
-# Now really mine LCDd.conf information
+# Now really mine LCDd.conf information using Dummy tree
 
 # loop over all INI classes
 foreach my $ini_class (@ini_classes) {
@@ -307,21 +345,21 @@ foreach my $ini_class (@ini_classes) {
     # Now create a an $ini_class element in LCDd class (to link LCDd
     # class and LCDd::$ini_class)
     my $driver_class_spec = qq!
-        class:LCDd 
-            element:$ini_class 
+        class:LCDd
+            element:$ini_class
     ! ;
 
     if ( $ini_class eq 'server' or $ini_class eq 'menu' ) {
-        $driver_class_spec .= qq! 
-            type=node 
-            config_class_name="LCDd::$ini_class" 
+        $driver_class_spec .= qq!
+            type=node
+            config_class_name="LCDd::$ini_class"
         ! ;
     }
     else {
         # Arrange a driver class is shown only if the driver was selected
         # in the [server] class
-        $driver_class_spec .= qq! 
-            type=warped_node 
+        $driver_class_spec .= qq!
+            type=warped_node
             config_class_name="LCDd::$ini_class"
             level=hidden
             warp
@@ -335,14 +373,14 @@ foreach my $ini_class (@ini_classes) {
 
 ######################
 #
-# Step3: write the model 
+# Step 5: write the model
 
 
 # Itself constructor returns an object to read or write the data
 # structure containing the model to be edited. force_write is required
 # because writer object, being created *after* loading the model in the
 # instance, is not aware of these changes.
-my $rw_obj = Config::Model::Itself->new( 
+my $rw_obj = Config::Model::Itself->new(
     model_object => $meta_root,
     cm_lib_dir => 'lib/Config/Model/',
     force_write => 1,
@@ -351,6 +389,9 @@ my $rw_obj = Config::Model::Itself->new(
 say "Writing all models in file (please wait)";
 $rw_obj->write_all;
 
+# mop up
+$tmp->remove_tree;
+
 say "Done";
 
 # this function extracts info specified between square brackets and returns a model snippet
@@ -361,7 +402,7 @@ sub info_to_model {
     my @model ;
 
     # legal needs to be parsed first to setup value_type first
-    my %info = map { split /[:=]/,$_ ,2 ; } split /;/,$info ; 
+    my %info = map { split /[:=]/,$_ ,2 ; } split /;/,$info ;
 
     # use this semantic information to better specify the parameter
     if (my $legal = delete $info{legal} || '') {
